        .MODEL tiny
        .386
        .CODE
        org 7c00h
        
; Sample VBE (Vesa Bios Extension) 'bootsector'
; Sets 800 x 600 with a linear frame buffer in either 24 or 32-bits directcolor
; The program then draws a red gradient background and draws text over (using an
; alpha mask so the background behind stays visible) it.
;
; Programmed by Rob Lohman (except for unreal mode code), visuar@iname.com
; March 15, 2006. Use this code anyway you want.
;
; Compiles with MASM 6.11 with the following command line:
; ml /AT /Fe Vesa.img Vesa.asm
;
; Requirements:
; - minimal 486 (for bswap)
; - graphics card with at least VESA 2.0 BIOS and 2 MB of onboard memory
; - fast A20 enabled
;
; For some reason this does not work with QEMU on my system (it does on Bochs!)
; It displays a garbled screen (even before the drawing happens) with the
; CIRRUS BIOS. Their VBE BIOS does not even return a mode list (strange enough).
; So that BIOS probably only works with the VESA "default" modes. Since Bochs
; uses the exact same BIOS (yes, I tried the latest one with QEMU) something
; most be wrong in QEMU probably.
;
; The code has been tested (to work) on the following systems:
; - DELL Laptop with Pentium 4, 1 GB of RAM, ATI Radeon Mobility 9700 128 MB AGP [24-bit mode]
; - AMD Athlon system, 768 MB of RAM, nVidia GeForce 2 MX400 64 MB AGP [32-bit mode]
; - Intel Pentium Celeron, 256 MB of RAM, S3 Savage3D 8MB AGP & nVidia Riva 128 4 MB [both 32-bit mode]
; - Microsoft Virtual PC 2004, 128 MB of RAM, IBM SVGA BIOS [32-bit mode], emulated
; - Bochs 2.2.6, 32 MB of RAM, Bochs/Plex86 VBE [24-bit mode]
; - QEMU isn't able to switch correctly to the graphics mode
;
; unreal mode code from: http://my.execpc.com/CE/AC/geezer/osd/pmode/unreal.asm
;
; To make sure everything fits in 512 bytes (all the operand / address prefixes
; for 32-bit usage in 16-bit mode is eating a lot of bytes and this is exactly
; 512 bytes long!) I had to cut some corners (see commented code throughout the file):
; - fast A20 enable is used
; - no checking if string will flow off the screen buffer (more than 7500 characters)
; - error messages are cryptic :)
;
; The program displays a 'V' character (in textmode) if it could not find
; VBE 2.0 or higher. It displays a 'M' character if it cannot find a
; suitable 800 x 600 video mode.

include Vesa.inc	; vesa structures & defines        

Start proc
        cli
        xor			ax, ax
        mov     ss, ax
        mov     sp, 7C00h					; setup stack to start below this code
        mov			ds, ax						; data segment starts at 0
        sti
        cld

				call		CheckForVBE				; if this returns we have correct VBE and GS has been setup
				
				call		FindAndSetMode		; if this returns suitable mode was found & set, mode info is in es:0000h
				
				call		EnableUnrealMode	; enable A20 line and setup unreal mode to allow access to the full 4 GB of memory
				
				; ***************************************************************************************************
				; 1. checks for VBE 2.0 have been done
				; 2. 800 x 600 in either 24 or 32 bits direct color with a linear frame buffer has been found and set
				; 3. A20 line and unreal mode have been enabled
				; ***************************************************************************************************
				
				; **** THE DRAWING STARTS HERE ****
				
				call		DrawBackground
				
				call		DrawCharacters		; no call to reduce space
				
				jmp			$ 								; hang				
Start endp

; **** draw video card vendor string over background ****
DrawCharacters proc
				mov			edi, ES:[VesaMode.ModePhysBasePtr]
				mov			esi, ebp					; secure bytes to skip per pixel (ebp still has that value)
				
				mov			eax, 8
				mov			cx, si
				mul			cl								; calculates bytes per character row (result in AX)

				push		edx								; row length (store it for later use)

				sub			edx, eax					; edx still contains the amount of bytes per row, we need that minus the length of one character row
				sub			edi, edx					; undo the first "add edi, edx" below
				push		edx								; store the row skip for later use
				
				; retrieve the 8 x 8 video rom character set
				mov			ax, 1130h
				mov			bh, 3							; get first 128 characters for 8 x 8 font (use 4 to get second half)
				int			10h								; ES:BP contains the pointer to the character set (destroys cx & dx)
																	; Each bit in the character map represents a pixel (so 8 bits [1 byte] is 1 character row, a full character is 8 bytes)
				mov			ax, 2000h
				mov			fs, ax						; set FS to VBE info structure
				
				mov			dx, FS:[VesaInfo.VesaVendorSegment]
				mov			bx, FS:[VesaInfo.VesaVendorOffset]
				mov			fs, dx						; FS:BX points to video card vendor string
				
CharacterLoop:
				push		bx				
				mov			bl, FS:[bx]				; get vendor string character
				
				test		bl, bl
				jz			DrawEnd						; stop when we are at the end of the string
				
				xor			bh, bh
				shl			bx, 3							; multiply by 8 (bytes [per character], to get index into character map)
				add			bx, bp						; add base of character map to the index
				
				mov			edx, ES:[bx]			; get rows 4 - 1 from the character cell
				mov			ebx, ES:[bx + 4]	; get rows 8 - 5 from the character cell
				.486
				bswap		edx								; 4 - 1 ==> 1 - 4
				bswap		ebx								; 8 - 5 ==> 5 - 8
				.386
				
				mov			cx, 64						; 64 bits (8 x 8) in a character cell
				
PlotCharacterPixel:
				test		cx, 7
				jnz			NotOnNewLine				
				
				push		edx
				mov			edx, [esp + 6]		; get the row skip from the stack (push edx above)
				add			edi, edx					; add (width * bytes_per_pixel) - 8
				pop			edx								; restore the second half of the character cell
				
NotOnNewLine:											
				xor			eax, eax					; start with fully transparant mask (eax = high bit edx:ebx)
				shld		eax, edx, 1				; shift the next pixel into eax (ebx does not change!)
				shld		edx, ebx, 1				; shift a new pixel from the second half into the first half of the character (edx does not change)
				shl			ebx, 1						; shift the second half (this is a NOP after half the character is drawn)
								
				not			eax								; 0x00000000 -> 0xFFFFFFFF, 0x00000001 -> 0xFFFFFFFE if this character has no pixel leave the
				inc			eax								; 0xFFFFFFFF -> 0x00000000, 0xFFFFFFFE -> 0xFFFFFFFF mask empty, otherwise set the whole mask
				and			eax, 00FFFFFFh		; the top byte is not used in 24/32-bit modes
				
				or			DS:[edi], eax			; write the pixel leaving the background intact
				
				add			edi, esi					; skip 3 or 4 bytes
				loop		PlotCharacterPixel
				
				pop			bx
				inc			bx								; go to the next character in the vendor string
				
				mov			ecx, [esp + 4]		; get line length
				shl			cx, 3							; multiply by 8 (we need to get back to the top of the row)
				sub			edi, ecx			

				mov			eax, 8
				mov			cx, si
				mul			cl								; calculates bytes per character row (result in AX)

				add			edi, eax					; add offset to next character
				
				jmp			CharacterLoop

DrawEnd:
				pop			bx							
				pop			edx
				pop			edx
				ret
DrawCharacters endp

; **** draw background ****
; draws a red "gradient" background, it might be
; more optimized to use a different 24 or 32
; bits routine. But that would take more space
DrawBackground proc
				mov			edi, ES:[VesaMode.ModePhysBasePtr]
				
				; calculate if we need to add 4 & 3200 bytes per pixel / row or 3 & 2400 bytes
				xor			eax, eax
				mov			al, ES:[VesaMode.ModeBitsPerPixel]
				mov			ebp, eax
				shr			bp, 3							; BP will either be 4 (32 bits) or 3 (24 bits)
				
				mov			eax, 800					; number of pixels per line
				mul			bp								; multiply by number of bytes per pixel (stores result in dx:ax [16 bit mode!])
				mov			edx, eax

				; Officially the following line needs a check. If VBE is 3.0 VesaMode.LinRedFieldPos should be used instead!!
				mov			cl, ES:[VesaMode.ModeRedFieldPos]
				
				mov			ebx, 800 * 600		; total number of pixels to fill
				xor			eax, eax
				
RepeatDraw:				
				mov			ax, di
				and			ax, 0FFh					; make sure it is in the range 0 - 255
				
				shl			eax, cl						; shift it into the right color byte (red in this case)
				mov			ds:[edi], eax			; we can't do 'stosd' for example since that doesn't work in unreal mode!				
				shr			eax, cl						; shift the color back for the next gradient round

				add			edi, ebp					; add 3 or 4 bytes to skip to the next pixel
				
				dec			ebx
				jnz			RepeatDraw

				ret
DrawBackground endp

; **** make sure VBE exists and is at least version 2.0 ****
CheckForVBE proc
				; the 512 / 256 byte buffer can be put at 2000:0000
				mov			ax, 2000h
				mov			es, ax
				
				;mov			ES:[VesaInfo.VesaSig], VBE2_SIGNATURE		; no need to do this for this test, see VBE spec!
				xor			di, di
				mov			ax, 4F00h
				int			10h

				; check if vesa bios was found				
				cmp			ax, 4Fh
				jne			NoVesa
				cmp			ES:[VesaInfo.VesaSig], VESA_SIGNATURE
				jne			NoVesa
				
				; we need at least version 2 of the bios (linear memory support)
				cmp			byte ptr ES:[VesaInfo.VesaVersion] + 1, 2
				jb			NoVesa
				
				; get the pointer to the mode list
				mov			si, ES:[VesaInfo.VesaModesOffset]
				mov			ax, ES:[VesaInfo.VesaModesSegment]
				mov			gs, ax
				ret

NoVesa:				
				mov			al, 'V'
				call		DisplayLetter
				
				jmp			$		; hang
CheckForVBE endp

; **** Find a suitable graphics mode and activate it ****
FindAndSetMode proc
				; create a new segment to put the mode info in, we still may
				; need the previous one since Vesa modes can be *inside* that buffer!!
				mov			ax, 2100h				
				mov			es, ax
				
				; walk through the modes and find the one we need:
				; 800 x 600 in either 24 or 32 bits with linear frame buffer & not interlaced
				
VesaModeLoop:
				mov			cx, GS:[si]
				cmp			cx, 0FFFFh
				je			NoModeFound		; if we are at the end of the list we didn't find a mode
				inc			dx
				
				;push		si						; VESA spec indicates registers might be destroyed

				mov			ax, 4F01h			; get mode info, es has already been setup by the code above
				xor			di, di
				int			10h
				
				;pop			si
				
				cmp			ax, 4Fh
				jne			VesaModeContinue
				
				; now check if this is the mode
				mov			ax, ES:[VesaMode.ModeAttributes]
				and 		ax, ATTRIBUTES
				cmp			ax, ATTRIBUTES	; supported, color, graphics & linear frame buffer
				jne			VesaModeContinue			
				
				cmp			ES:[VesaMode.ModeHorizontalRes], 800	; 800 x 600
				jne			VesaModeContinue
				cmp			ES:[VesaMode.ModeVerticalRes], 600
				jne			VesaModeContinue
				
				cmp			ES:[VesaMode.ModeMemoryModel], MODE_DIRECT_COLOR	; only direct color modes (ie, no palette)
				jne			VesaModeContinue
				
				cmp			ES:[VesaMode.ModePhysBasePtr], 0	; mode must have physical base address
				je			VesaModeContinue
				
				cmp			ES:[VesaMode.ModeBitsPerPixel], 24
				je			FoundMode
				cmp			ES:[VesaMode.ModeBitsPerPixel], 32
				je			FoundMode

VesaModeContinue:
				inc			si
				inc			si
				jmp			VesaModeLoop
				
				; we found a suitable vesa mode, information about the mode is in es:0000h
				; and the mode number is in GS:[si]
FoundMode:
				mov			bx, GS:[si]			; get mode number
				or			bx, SET_LINEAR	; indicate we want linear mode
				mov			ax, 4F02h
				int			10h							; set the new mode
				
				cmp			ax, 4Fh
				jne			VesaModeContinue	; continue searching if we could not set this mode
				
				ret

NoModeFound:		
				mov			al, 'M'
				call		DisplayLetter
				
				jmp			$ ; hang
FindAndSetMode endp

; **** do a fast A20 enable and switch to unreal mode ****
EnableUnrealMode proc
				; do a fast enable A20 to make sure we can access all memory
				in 			al, 92h
        or 			al, 2
        out 		92h, al
				
				; switch to "UNREAL" mode to draw some pixels		
				
				.386P
				push 		ds
        
        cli             				; interrupts off
        lgdt 		fword ptr [gdt_ptr]
        
				mov 		eax, cr0
				or 			al, 1
        mov 		cr0, eax     		; partial switch to 32-bit pmode
        
        mov 		bx, 8				 		; selector to segment w/ 4G limit
				mov 		ds, bx					; set seg limits in descriptor caches
				
				dec 		al
        mov 		cr0, eax     		; back to (un)real mode
        
        pop 		ds              ; but now 32-bit addresses are OK

				ret
EnableUnrealMode endp

; ***** display routines for the (error) messages *****

DisplayLetter proc
				mov		bx, 7
				mov		ah, 0Eh
				int		10h
				ret
DisplayLetter endp

align
; ***** unreal mode gdt stuff *****
gdt				dw 	0                    ; limit 15:0
        	dw 	0                    ; base 15:0
        	db 	0                    ; base 23:16
        	db 	0                    ; access byte (descriptor type)
        	db 	0                    ; limit 19:16, flags
        	db 	0                    ; base 31:24

					dw 	0FFFFh
					dw 	0
					db 	0
        	db 	92h          				; present, ring 0, data, expand-up, writable
					db 	0CFh									; page-granular, 32-bit
					db 	0
gdt_end:

gdt_ptr 	dw 	gdt_end - gdt - 1    ; GDT limit
gdt_addr	dd 	offset gdt

; ***** magic marker *****

				org   7DFEh
				db		55h, 0AAh
				
				END Start